/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.alfa;

import java.util.Collection;
import java.util.Collections;
import java.util.Map;
import java.util.Objects;
import java.util.Optional;
import java.util.Set;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import org.apache.commons.collections.MapUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.core.convert.converter.Converter;
import org.springframework.lang.NonNull;
import org.springframework.security.authentication.AbstractAuthenticationToken;
import org.springframework.security.core.GrantedAuthority;
import org.springframework.security.core.authority.SimpleGrantedAuthority;
import org.springframework.security.oauth2.jwt.Jwt;
import org.springframework.security.oauth2.jwt.JwtClaimNames;
import org.springframework.security.oauth2.server.resource.authentication.JwtAuthenticationToken;
import org.springframework.security.oauth2.server.resource.authentication.JwtGrantedAuthoritiesConverter;
import org.springframework.stereotype.Component;

@Component
public class JwtAuthConverter implements Converter<Jwt, AbstractAuthenticationToken> {

	private final JwtGrantedAuthoritiesConverter jwtGrantedAuthoritiesConverter = new JwtGrantedAuthoritiesConverter();

	private static final String SIMPLE_GRANT_AUTHORITY_PREFIX = "ROLE_";
	static final String RESOURCE_ACCESS_KEY = "resource_access";
	static final String ROLES_KEY = "roles";

	@Autowired
	private OAuth2Properties oAuth2Properties;

	@Override
	public AbstractAuthenticationToken convert(@NonNull Jwt jwt) {
		return new JwtAuthenticationToken(jwt, getAuthorities(jwt), getPrincipleClaimName(jwt));
	}

	String getPrincipleClaimName(Jwt jwt) {
		return MapUtils.getString(jwt.getClaims(), getClaimName());
	}

	private String getClaimName() {
		return Optional.ofNullable(oAuth2Properties.getPrincipleAttribute()).orElse(JwtClaimNames.SUB);
	}

	Set<GrantedAuthority> getAuthorities(Jwt jwt) {
		return Stream.concat(jwtGrantedAuthoritiesConverter.convert(jwt).stream(), extractResourceRoles(jwt).stream()).collect(Collectors.toSet());
	}

	Collection<SimpleGrantedAuthority> extractResourceRoles(Jwt jwt) {
		var resourceAccess = getResourceAccess(jwt);
		if (resourceAccess.isEmpty()) {
			return Collections.emptySet();
		}
		if (Objects.isNull(resourceAccess.get(oAuth2Properties.getResource()))) {
			return Collections.emptySet();
		}
		return extractRoles(getClaimMapFromMap(resourceAccess, oAuth2Properties.getResource()));
	}

	private Map<String, Object> getResourceAccess(Jwt jwt) {
		return jwt.getClaimAsMap(RESOURCE_ACCESS_KEY);
	}

	@SuppressWarnings("unchecked")
	private Map<String, Object> getClaimMapFromMap(Map<String, Object> claimMap, String claimName) {
		return MapUtils.getMap(claimMap, claimName);
	}

	Set<SimpleGrantedAuthority> extractRoles(Map<String, Object> resource) {
		if (!resource.containsKey(ROLES_KEY)) {
			return Collections.emptySet();
		}
		return getRoles(resource).stream().map(role -> new SimpleGrantedAuthority(SIMPLE_GRANT_AUTHORITY_PREFIX + role)).collect(Collectors.toSet());
	}

	@SuppressWarnings("unchecked")
	private Collection<String> getRoles(Map<String, Object> resource) {
		return (Collection<String>) resource.get(ROLES_KEY);
	}
}