/*
 * Copyright (C) 2025 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.alfa.wiedervorlage;

import java.util.Map;
import java.util.Optional;
import java.util.stream.Stream;

import org.apache.commons.collections.MapUtils;
import org.springframework.stereotype.Component;

import de.ozgcloud.alfa.common.command.Command;
import de.ozgcloud.alfa.common.command.CommandBodyMapper;
import de.ozgcloud.alfa.common.command.CommandOrder;
import de.ozgcloud.alfa.common.command.LegacyOrder;
import de.ozgcloud.alfa.historie.HistorieProcessor;

@Component
class WiedervorlageHistorieProcessor implements HistorieProcessor {

	@Override
	public Stream<Command> process(Stream<Command> commands) {
		return commands
				.map(this::translateCreateAttachedItem)
				.map(this::translateUpdateAttachedItem)
				.map(this::translatePatchAttachedItem);
	}

	Command translateCreateAttachedItem(Command command) {
		if (isCreateWiederVorlage(command)) {
			return command.toBuilder().order(LegacyOrder.CREATE_WIEDERVORLAGE).build();
		}
		return command;
	}

	private boolean isCreateWiederVorlage(Command command) {
		return CommandOrder.CREATE_ATTACHED_ITEM == command.getCommandOrder()
				&& isWiedervorlageItem(command);
	}

	Command translateUpdateAttachedItem(Command command) {
		if (isEditWiedervorlage(command)) {
			return command.toBuilder().order(LegacyOrder.EDIT_WIEDERVORLAGE).build();
		}
		return command;
	}

	private boolean isEditWiedervorlage(Command command) {
		return CommandOrder.UPDATE_ATTACHED_ITEM == command.getCommandOrder()
				&& isWiedervorlageItem(command);
	}

	Command translatePatchAttachedItem(Command command) {
		if (isPatchWiedervorlage(command)) {
			return translatePatchWiedervorlage(command);
		}
		return command;
	}

	private boolean isPatchWiedervorlage(Command command) {
		return CommandOrder.PATCH_ATTACHED_ITEM == command.getCommandOrder() && isWiedervorlageItem(command);
	}

	private boolean isWiedervorlageItem(Command command) {
		return getItemName(command).filter(name -> name.equals(Wiedervorlage.class.getSimpleName())).isPresent();
	}

	private Optional<String> getItemName(Command command) {
		return Optional.ofNullable(MapUtils.getString(command.getBody(), CommandBodyMapper.ITEM_NAME_PROPERTY));
	}

	Command translatePatchWiedervorlage(Command command) {
		return getDoneValue(command).map(isDone -> resolvePatchWiedervorlage(command, isDone)).orElse(command);

	}

	private Optional<Boolean> getDoneValue(Command command) {
		return Optional.ofNullable(MapUtils.getBoolean(getItemMap(command), Wiedervorlage.DONE_FIELD));
	}

	@SuppressWarnings("unchecked")
	private Map<String, Object> getItemMap(Command command) {
		return MapUtils.getMap(command.getBody(), CommandBodyMapper.ITEM_PROPERTY);
	}

	private Command resolvePatchWiedervorlage(Command command, Boolean isDone) {
		return isDone.booleanValue() ? command.toBuilder().order(LegacyOrder.WIEDERVORLAGE_ERLEDIGEN).build()
				: command.toBuilder().order(LegacyOrder.WIEDERVORLAGE_WIEDEREROEFFNEN).build();
	}
}
