/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.alfa;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.util.Optional;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;
import org.springframework.hateoas.IanaLinkRelations;
import org.springframework.hateoas.Link;
import org.springframework.test.util.ReflectionTestUtils;

import de.ozgcloud.alfa.common.user.CurrentUserService;
import de.ozgcloud.alfa.common.user.UserManagerUrlProvider;
import de.ozgcloud.alfa.common.user.UserProfileTestFactory;
import de.ozgcloud.alfa.common.user.UserRole;
import de.ozgcloud.alfa.system.SystemStatusService;

class RootModelAssemblerTest {

	@Spy
	@InjectMocks
	private RootModelAssembler modelAssembler;
	@Mock
	private CurrentUserService currentUserService;
	@Mock
	private SystemStatusService systemStatusService;
	@Mock
	private UserManagerUrlProvider userManagerUrlProvider;
	@Mock
	private RootViewLinkHandler viewLinkHandler;

	private Root root = RootTestFactory.create();

	@DisplayName("Links for user")
	@Nested
	class TestLinks {

		@DisplayName("with role " + UserRole.VERWALTUNG_USER)
		@Nested
		class TestWithVerwaltungRole {

			@BeforeEach
			void mockCurrentUserService() {
				doReturn(true).when(modelAssembler).hasVerwaltungRole();
			}

			@Test
			void shouldHaveSelfLink() {
				var model = modelAssembler.toModel(root);

				assertThat(model.getLink(IanaLinkRelations.SELF)).isPresent();
			}

			@Test
			void shouldHaveDownloadTokenLink() {
				var model = modelAssembler.toModel(root);

				assertThat(model.getLink(RootModelAssembler.REL_DOWNLOAD_TOKEN)).isPresent().get().extracting(Link::getHref)
						.isEqualTo("/api/downloadtoken");
			}
		}

		@DisplayName("with no role")
		@Nested
		class TestWithNoRole {

			@BeforeEach
			void mockCurrentUserService() {
				when(currentUserService.hasRole(anyString())).thenReturn(false);
			}

			@Test
			void shouldHaveNoLinkAtAll() {
				var model = modelAssembler.toModel(root);

				assertThat(model.getLinks()).isEmpty();
			}
		}

		@DisplayName("current user")
		@Nested
		class CurrentUser {

			@DisplayName("when usermanager url is configured")
			@Nested
			class UsermanagerUrlConfigured {

				private String userProfileTemplate = "UserProfileTemplate/%s";

				@BeforeEach
				void mock() {
					when(userManagerUrlProvider.isConfiguredForUserProfile()).thenReturn(true);
					when(userManagerUrlProvider.getUserProfileTemplate()).thenReturn(userProfileTemplate);
					when(currentUserService.findUserId()).thenReturn(Optional.of(UserProfileTestFactory.ID));
				}

				@Test
				void shouldHaveCurrentUserLinkWithUserId() {
					var model = modelAssembler.toModel(root);

					assertThat(model.getLink(RootModelAssembler.REL_CURRENT_USER))
							.isPresent().get().extracting(Link::getHref)
							.isEqualTo("UserProfileTemplate/" + UserProfileTestFactory.ID);
				}

			}

			@DisplayName("when usermanager url is not configured")
			@Nested
			class UsermanagerUrlNotConfigured {

				@BeforeEach
				void mock() {
					when(userManagerUrlProvider.isConfiguredForUserProfile()).thenReturn(false);
				}

				@Test
				void shouldNotHaveCurrentUserLink() {
					var model = modelAssembler.toModel(root);

					assertThat(model.getLink(RootModelAssembler.REL_CURRENT_USER)).isNotPresent();
				}
			}
		}
	}

	@DisplayName("Test user assistance documentation link")
	@Nested
	class TestDocumentationLink {

		private static final String DOCUMENTATION_URL = "http://alfa-docs";

		@Test
		void shouldNotHaveLinkForUndefinedUrl() {
			var link = modelAssembler.toModel(root);

			assertThat(link.getLink(RootModelAssembler.REL_DOCUMENTATIONS)).isEmpty();
		}

		@Test
		void shouldHaveLinkForDefinedUrl() {
			ReflectionTestUtils.setField(modelAssembler, "documentationUrl", DOCUMENTATION_URL);

			var link = modelAssembler.toModel(root);

			assertThat(link.getLink(RootModelAssembler.REL_DOCUMENTATIONS))
					.isPresent()
					.map(Link::getHref)
					.hasValue(DOCUMENTATION_URL);
		}

		@Test
		void shouldNotHaveLinkForEmptyStringUrl() {
			ReflectionTestUtils.setField(modelAssembler, "documentationUrl", "");

			var link = modelAssembler.toModel(root);

			assertThat(link.getLink(RootModelAssembler.REL_DOCUMENTATIONS)).isEmpty();
		}
	}
}
