/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.alfa;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.util.List;
import java.util.Optional;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.EnumSource;
import org.junit.jupiter.params.provider.EnumSource.Mode;
import org.mockito.AdditionalAnswers;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;
import org.springframework.hateoas.EntityModel;
import org.springframework.hateoas.Link;
import org.springframework.hateoas.LinkRelation;

import com.thedeanda.lorem.LoremIpsum;

import de.ozgcloud.alfa.common.FeatureToggleProperties;
import de.ozgcloud.alfa.common.user.CurrentUserService;
import de.ozgcloud.alfa.common.user.UserId;
import de.ozgcloud.alfa.common.user.UserProfileTestFactory;
import de.ozgcloud.alfa.common.user.UserRole;
import de.ozgcloud.alfa.system.SystemStatusService;
import de.ozgcloud.alfa.vorgang.Vorgang.VorgangStatus;

class RootViewLinkHandlerTest {

	@Spy
	@InjectMocks
	private RootViewLinkHandler viewLinkHandler;
	@Mock
	private CurrentUserService currentUserService;
	@Mock
	private SystemStatusService systemStatusService;
	@Mock
	private FeatureToggleProperties featureToggleProperties;

	@DisplayName("Add view links")
	@Nested
	class TestAddViewLinks {

		private Root rootResource = RootTestFactory.create();

		private Optional<UserId> emptyUserId = Optional.empty();

		@Test
		void shouldAddVorgaengeAllLink() {
			var model = EntityModel.of(rootResource);

			viewLinkHandler.addViewLinks(model, emptyUserId);

			assertThat(model.getLink(RootViewLinkHandler.ALL_VORGAENGE_REL))
					.get().extracting(Link::getHref).isEqualTo("/api/vorgangs?page=0&limit=100");
		}

		@Test
		void shouldCallAddSearchAllVorgaengeLink() {
			var model = EntityModel.of(rootResource);

			viewLinkHandler.addViewLinks(model, emptyUserId);

			verify(viewLinkHandler).addSearchAllVorgaengeLink(model);
		}

		@Test
		void shouldAddLinksForVerwaltungPoststelle() {
			when(currentUserService.hasRole(UserRole.VERWALTUNG_POSTSTELLE)).thenReturn(true);

			var model = EntityModel.of(rootResource);
			viewLinkHandler.addViewLinks(model, emptyUserId);

			verify(viewLinkHandler).addViewLinksForVerwaltungPoststelle(model, emptyUserId);
		}

		@Test
		void shouldAddLinksForVerwaltungUser() {
			when(currentUserService.hasRole(UserRole.VERWALTUNG_POSTSTELLE)).thenReturn(false);
			when(currentUserService.hasRole(UserRole.VERWALTUNG_USER)).thenReturn(true);

			var model = EntityModel.of(rootResource);
			viewLinkHandler.addViewLinks(model, emptyUserId);

			verify(viewLinkHandler).addViewLinksForVerwaltungUser(model, emptyUserId);
		}

		@DisplayName("for " + UserRole.VERWALTUNG_POSTSTELLE)
		@Nested
		class TestVerwaltungPoststelle {

			private final EntityModel<Root> model = EntityModel.of(rootResource);

			@Test
			void shouldCallBuildMyVorgaengeLink() {
				viewLinkHandler.addViewLinksForVerwaltungPoststelle(model, Optional.of(UserProfileTestFactory.ID));

				verify(viewLinkHandler).buildMyVorgaengeLink(UserProfileTestFactory.ID);
			}

			@DisplayName("search my vorgaenge")
			@Nested
			class TestSearchMyVorgaenge {

				@DisplayName("on configured search server")
				@Nested
				class TestOnConfiguredSearchServer {

					@Test
					void shouldCallBuildSearchMyVorgaengeLink() {
						when(systemStatusService.isSearchServerAvailable()).thenReturn(true);

						viewLinkHandler.addViewLinksForVerwaltungPoststelle(model, Optional.of(UserProfileTestFactory.ID));

						verify(viewLinkHandler).buildSearchMyVorgaengeLink(UserProfileTestFactory.ID);
					}
				}

				@DisplayName("on not configured search server")
				@Nested
				class TestOnNotConfiguredSearchServer {

					@Test
					void shouldNotCallBuildSearchMyVorgaengeLink() {
						when(systemStatusService.isSearchServerAvailable()).thenReturn(false);

						viewLinkHandler.addViewLinksForVerwaltungPoststelle(model, Optional.of(UserProfileTestFactory.ID));

						verify(viewLinkHandler, never()).buildSearchMyVorgaengeLink(any());
					}
				}
			}
		}

		@DisplayName("for " + UserRole.VERWALTUNG_USER)
		@Nested
		class TestVerwaltungUser {

			private final EntityModel<Root> model = EntityModel.of(rootResource);

			@Test
			void shouldCallAddUnassignedVorgaengeLink() {
				viewLinkHandler.addViewLinksForVerwaltungUser(model, Optional.of(UserProfileTestFactory.ID));

				verify(viewLinkHandler).buildGetAllUnassignedVorgaengeLink();
			}

			@Test
			void shouldCallAddMyVorgaengeLink() {
				viewLinkHandler.addViewLinksForVerwaltungUser(model, Optional.of(UserProfileTestFactory.ID));

				verify(viewLinkHandler).buildMyVorgaengeLink(UserProfileTestFactory.ID);
			}

			@DisplayName("search my vorgaenge")
			@Nested
			class TestSearchMy {

				@DisplayName("on configured search server")
				@Nested
				class TestOnConfiguredSearchServer {

					@Test
					void shouldCallBuildSearchMyVorgaengeLink() {
						when(systemStatusService.isSearchServerAvailable()).thenReturn(true);

						viewLinkHandler.addViewLinksForVerwaltungUser(model, Optional.of(UserProfileTestFactory.ID));

						verify(viewLinkHandler).buildSearchMyVorgaengeLink(UserProfileTestFactory.ID);
					}
				}

				@DisplayName("on not configured search server")
				@Nested
				class TestOnNotConfiguredSearchServer {

					@Test
					void shouldNotCallBuildSearchMyVorgaengeLink() {
						when(systemStatusService.isSearchServerAvailable()).thenReturn(false);

						viewLinkHandler.addViewLinksForVerwaltungUser(model, Optional.of(UserProfileTestFactory.ID));

						verify(viewLinkHandler, never()).buildSearchMyVorgaengeLink(any());
					}
				}
			}

			@Test
			void shouldCallAddAllVorgangStatusLinks() {
				viewLinkHandler.addViewLinksForVerwaltungUser(model, Optional.of(UserProfileTestFactory.ID));

				verify(viewLinkHandler).addAllVorgangStatusLinks(model);
			}

			@Test
			void shouldCallAddMyVorgaengeStatusLinks() {
				viewLinkHandler.addViewLinksForVerwaltungUser(model, Optional.of(UserProfileTestFactory.ID));

				verify(viewLinkHandler).buildMyVorgaengeStatusLinks(UserProfileTestFactory.ID);
			}

			@Test
			void shouldCallAddUnassignedVorgaengeStatusLinks() {
				viewLinkHandler.addViewLinksForVerwaltungUser(model, Optional.of(UserProfileTestFactory.ID));

				verify(viewLinkHandler).buildUnassignedVorgaengeStatusLinks();
			}

			@DisplayName("search unassigned vorgaenge")
			@Nested
			class TestSearchUnassignedVorgaengeLink {

				@DisplayName("with configured search server")
				@Nested
				class TestWithConfiguredSearchServer {

					@Test
					void shouldCallBuildSearchUnassignedVorgaengeLink() {
						when(systemStatusService.isSearchServerAvailable()).thenReturn(true);

						viewLinkHandler.addViewLinksForVerwaltungUser(model, Optional.of(UserProfileTestFactory.ID));

						verify(viewLinkHandler).buildSearchUnassignedVorgaengeLink();
					}
				}

				@DisplayName("with not configured search server")
				@Nested
				class TestWithNotConfiguredSearchServer {

					@Test
					void shouldNotCallBuildSearchUnassignedVorgaengeLink() {
						when(systemStatusService.isSearchServerAvailable()).thenReturn(false);

						viewLinkHandler.addViewLinksForVerwaltungUser(model, Optional.of(UserProfileTestFactory.ID));

						verify(viewLinkHandler, never()).buildSearchUnassignedVorgaengeLink();
					}
				}
			}

			@DisplayName("vorgaenge with open wiedervorlagen")
			@Nested
			class TestVorgaengeWithOpenWiedervorlagen {

				@Test
				void shouldAddAllLink() {
					var model = EntityModel.of(rootResource);
					viewLinkHandler.addViewLinksForVerwaltungUser(model, Optional.of(UserProfileTestFactory.ID));

					assertThat(model.getLink(RootViewLinkHandler.ALL_WIEDERVORLAGEN_REL))
							.get().extracting(Link::getHref).isEqualTo("/api/vorgangs?page=0&limit=100&nextFrist=exists");
				}

				@DisplayName("considering user")
				@Nested
				class TestConsideringUser {

					@Test
					void shouldAddAssignedLinkIfExists() {
						var model = EntityModel.of(rootResource);
						viewLinkHandler.addViewLinksForVerwaltungUser(model, Optional.of(UserProfileTestFactory.ID));

						assertThat(model.getLink(RootViewLinkHandler.MY_WIEDERVORLAGEN_REL))
								.get().extracting(Link::getHref)
								.isEqualTo("/api/vorgangs?page=0&limit=100&assignedTo=" + UserProfileTestFactory.ID + "&nextFrist=exists");
					}

					@Test
					void shoulNotAddAssignedLinkIfNotExists() {
						var model = EntityModel.of(rootResource);
						viewLinkHandler.addViewLinksForVerwaltungUser(model, Optional.empty());

						assertThat(model.getLink(RootViewLinkHandler.MY_WIEDERVORLAGEN_REL)).isNotPresent();
					}
				}

				@Test
				void shouldAddUnassignedLink() {
					var model = EntityModel.of(rootResource);
					viewLinkHandler.addViewLinksForVerwaltungUser(model, Optional.of(UserProfileTestFactory.ID));

					assertThat(model.getLink(RootViewLinkHandler.UNASSIGNED_WIEDERVORLAGEN_REL))
							.get().extracting(Link::getHref).isEqualTo("/api/vorgangs?page=0&limit=100&assignedTo=&nextFrist=exists");
				}
			}

			@Test
			void shouldAddLinksForUngeleseneNachrichten() {
				viewLinkHandler.addViewLinksForVerwaltungUser(model, Optional.of(UserProfileTestFactory.ID));

				verify(viewLinkHandler).addGetByUngeleseneNachrichtenLinks(model, Optional.of(UserProfileTestFactory.ID));
			}
		}

		@DisplayName("search all vorgaenge")
		@Nested
		class TestSearchAllVorgaenge {

			@Test
			void shouldBeAddedIfSearchServerIsAvailable() {
				when(systemStatusService.isSearchServerAvailable()).thenReturn(true);

				var model = EntityModel.of(rootResource);
				viewLinkHandler.addSearchAllVorgaengeLink(model);

				assertThat(model.getLink(RootViewLinkHandler.SEARCH_ALL_REL))
						.get().extracting(Link::getHref).isEqualTo("/api/vorgangs?page=0&limit=100&searchBy={searchBy}");
			}

			@Test
			void shouldNotBeAddedIfSearchServerIsUnavailable() {
				when(systemStatusService.isSearchServerAvailable()).thenReturn(false);

				var model = spy(EntityModel.of(rootResource));
				viewLinkHandler.addSearchAllVorgaengeLink(model);

				verify(model, never()).add(any(Link.class));
			}
		}

		@DisplayName("my vorgaenge")
		@Nested
		class TestAddMyLink {

			@Test
			void shouldBeAddedIfUserIsPresent() {
				var link = viewLinkHandler.buildMyVorgaengeLink(UserProfileTestFactory.ID);

				assertThat(link.getHref()).isEqualTo("/api/vorgangs?page=0&limit=100&assignedTo=" + UserProfileTestFactory.ID);
			}
		}

		@DisplayName("search my vorgaenge")
		@Nested
		class TestSearchMyVorgaengeLink {

			@Test
			void shouldCallBuild() {
				viewLinkHandler.buildSearchMyVorgaengeLink(UserProfileTestFactory.ID);

				verify(viewLinkHandler).buildGetAllVorgaengeByAssignedToAndSearchByLink(UserProfileTestFactory.ID);
			}

			@Test
			void shouldBuildLinkWithMyVorgaengeRel() {
				var link = viewLinkHandler.buildSearchMyVorgaengeLink(UserProfileTestFactory.ID);

				assertThat(link.getRel()).extracting(LinkRelation::value).isEqualTo(RootViewLinkHandler.SEARCH_MY_VORGEANGE_REL);
			}
		}

		@DisplayName("search unassigned vorgaenge")
		@Nested
		class TestSearchUnassignedVorgaenge {

			@Test
			void shouldBuildLink() {
				var link = viewLinkHandler.buildSearchUnassignedVorgaengeLink();

				assertThat(link.getHref()).isEqualTo("/api/vorgangs?page=0&limit=100&assignedTo=&searchBy={searchBy}");
				assertThat(link.getRel().value()).isEqualTo(RootViewLinkHandler.SEARCH_UNASSIGNED_VORGEANGE_REL);
			}
		}

		@DisplayName("search all vorgaenge by assignedTo")
		@Nested
		class TestBuildGetAllVorgaengeByAssigendToAndSearchLink {

			@Test
			void shouldBuildLink() {
				var link = viewLinkHandler.buildGetAllVorgaengeByAssignedToAndSearchByLink(UserProfileTestFactory.ID).withSelfRel();

				assertThat(link).extracting(Link::getHref)
						.isEqualTo("/api/vorgangs?page=0&limit=100&assignedTo=" + UserProfileTestFactory.ID.toString() + "&searchBy={searchBy}");
			}
		}

		@DisplayName("all vorgang status")
		@Nested
		class TestAddAllVorgangStatusLinks {

			private final EntityModel<Root> model = EntityModel.of(rootResource);

			@Test
			void shouldCallAddVorgangStatusLinks() {
				viewLinkHandler.addAllVorgangStatusLinks(model);

				verify(viewLinkHandler).addVorgangStatusLinks(model, RootViewLinkHandler.VORGAENGE_ALL_REL_TEMPLATE);
			}
		}

		@DisplayName("my vorgang status")
		@Nested
		class TestAddMyVorgangStatusLinks {

			@Test
			void shouldCallAddVorgangStatusLink() {
				viewLinkHandler.buildMyVorgaengeStatusLinks(UserProfileTestFactory.ID);

				verify(viewLinkHandler).buildVorgangStatusLinks(RootViewLinkHandler.VORGAENGE_MY_REL_TEMPLATE, UserProfileTestFactory.ID);
			}

		}

		@DisplayName("vorgang status")
		@Nested
		class TestBuildVorgangStatusLinks {

			private final UserId userId = UserProfileTestFactory.ID;
			private final List<VorgangStatus> statuses = List.of(VorgangStatus.NEU, VorgangStatus.ANGENOMMEN);

			@BeforeEach
			void mock() {
				doReturn(statuses.stream()).when(viewLinkHandler).getVorgangStatuses();
			}

			@Test
			void shouldCallGetVorgangStatuses() {
				viewLinkHandler.buildVorgangStatusLinks(RootViewLinkHandler.VORGAENGE_ALL_REL_TEMPLATE, UserProfileTestFactory.ID);

				verify(viewLinkHandler).getVorgangStatuses();
			}

			@Test
			void shouldCallBuildVorgangListByPageLink() {
				viewLinkHandler.buildVorgangStatusLinks(RootViewLinkHandler.VORGAENGE_ALL_REL_TEMPLATE, UserProfileTestFactory.ID);

				statuses.forEach(status -> verify(viewLinkHandler).buildGetAllVorgaengeByAssignedToAndStatus(status,
						RootViewLinkHandler.VORGAENGE_ALL_REL_TEMPLATE, userId));
			}

			@Test
			void shouldReturnLinks() {
				var expectedLinks = LinkTestFactory.createList(statuses.size());
				doAnswer(AdditionalAnswers.returnsElementsOf(expectedLinks)).when(viewLinkHandler).buildGetAllVorgaengeByAssignedToAndStatus(any(),
						any(), any());

				var links = viewLinkHandler.buildVorgangStatusLinks(RootViewLinkHandler.VORGAENGE_ALL_REL_TEMPLATE, UserProfileTestFactory.ID);

				assertThat(links).containsExactlyInAnyOrderElementsOf(expectedLinks);
			}
		}

		@DisplayName("build get all vorgaenge by assignedTo and status")
		@Nested
		class TestBuildGetAllVorgaengeByAssignedToAndStatus {

			@ParameterizedTest
			@EnumSource(mode = Mode.INCLUDE, names = { "NEU", "ANGENOMMEN", "IN_BEARBEITUNG", "BESCHIEDEN", "ABGESCHLOSSEN", "VERWORFEN",
					"ZU_LOESCHEN", "WEITERGELEITET" })
			void shouldCallFormatLinkRelTemplate(VorgangStatus status) {
				viewLinkHandler.buildGetAllVorgaengeByAssignedToAndStatus(status, RootViewLinkHandler.VORGAENGE_MY_REL_TEMPLATE,
						UserProfileTestFactory.ID);

				verify(viewLinkHandler).formatLinkRelTemplate(RootViewLinkHandler.VORGAENGE_MY_REL_TEMPLATE, status);
			}

			@ParameterizedTest
			@EnumSource(mode = Mode.INCLUDE, names = { "NEU", "ANGENOMMEN", "IN_BEARBEITUNG", "BESCHIEDEN", "ABGESCHLOSSEN", "VERWORFEN",
					"ZU_LOESCHEN", "WEITERGELEITET" })
			void shouldAddByAssignedToAndStatus(VorgangStatus status) {
				var link = viewLinkHandler.buildGetAllVorgaengeByAssignedToAndStatus(status, RootViewLinkHandler.VORGAENGE_MY_REL_TEMPLATE,
						UserProfileTestFactory.ID);

				assertThat(link.getHref()).isEqualTo(
						"/api/vorgangs?page=0&limit=100&assignedTo=" + UserProfileTestFactory.ID.toString() + "&status=" + status.toString());
			}

			@ParameterizedTest
			@EnumSource(mode = Mode.INCLUDE, names = { "NEU", "ANGENOMMEN", "IN_BEARBEITUNG", "BESCHIEDEN", "ABGESCHLOSSEN", "VERWORFEN",
					"ZU_LOESCHEN", "WEITERGELEITET" })
			void shouldHaveFormattedRelation(VorgangStatus status) {
				var expectedRelation = LoremIpsum.getInstance().getWords(1);
				doReturn(expectedRelation).when(viewLinkHandler).formatLinkRelTemplate(any(), any());

				var link = viewLinkHandler.buildGetAllVorgaengeByAssignedToAndStatus(status, RootViewLinkHandler.VORGAENGE_MY_REL_TEMPLATE,
						UserProfileTestFactory.ID);

				assertThat(link.getRel()).isEqualTo(LinkRelation.of(expectedRelation));
			}
		}

		@DisplayName("format linkrel template")
		@Nested
		class TestFormatLinkRelTemplate {

			@ParameterizedTest
			@EnumSource(mode = Mode.INCLUDE, names = { "NEU", "ANGENOMMEN", "IN_BEARBEITUNG", "BESCHIEDEN", "ABGESCHLOSSEN", "VERWORFEN",
					"ZU_LOESCHEN", "WEITERGELEITET" })
			void shouldCallFormatLinkRelTemplateWithAllVorgaengeTemplate(VorgangStatus status) {
				var linkRel = viewLinkHandler.formatLinkRelTemplate(RootViewLinkHandler.VORGAENGE_ALL_REL_TEMPLATE, status);

				assertThat(linkRel).isEqualTo("vorgaenge_" + status.toString().toLowerCase() + "_all");
			}

			@ParameterizedTest
			@EnumSource(mode = Mode.INCLUDE, names = { "NEU", "ANGENOMMEN", "IN_BEARBEITUNG", "BESCHIEDEN", "ABGESCHLOSSEN", "VERWORFEN",
					"ZU_LOESCHEN", "WEITERGELEITET" })
			void shouldCallFormatLinkRelTemplateWithMyVorgaengeTemplate(VorgangStatus status) {
				var linkRel = viewLinkHandler.formatLinkRelTemplate(RootViewLinkHandler.VORGAENGE_MY_REL_TEMPLATE, status);

				assertThat(linkRel).isEqualTo("vorgaenge_" + status.toString().toLowerCase() + "_my");
			}

			@ParameterizedTest
			@EnumSource(mode = Mode.INCLUDE, names = { "NEU", "ANGENOMMEN", "IN_BEARBEITUNG", "BESCHIEDEN", "ABGESCHLOSSEN", "VERWORFEN",
					"ZU_LOESCHEN", "WEITERGELEITET" })
			void shouldCallFormatLinkRelTemplateWithUnassignedVorgaengeTemplate(VorgangStatus status) {
				var linkRel = viewLinkHandler.formatLinkRelTemplate(RootViewLinkHandler.VORGAENGE_UNASSIGNED_REL_TEMPLATE, status);

				assertThat(linkRel).isEqualTo("vorgaenge_" + status.toString().toLowerCase() + "_unassigned");
			}
		}

		@DisplayName("build get all vorgaenge by status")
		@Nested
		class TestBuildGetAllVorgaengeByStatus {

			@ParameterizedTest
			@EnumSource(mode = Mode.INCLUDE, names = { "NEU", "ANGENOMMEN", "IN_BEARBEITUNG", "BESCHIEDEN", "ABGESCHLOSSEN", "VERWORFEN",
					"ZU_LOESCHEN", "WEITERGELEITET" })
			void shouldAddByStatus(VorgangStatus status) {
				var link = viewLinkHandler.buildGetAllVorgaengeByStatus(status, RootViewLinkHandler.VORGAENGE_MY_REL_TEMPLATE);

				assertThat(link.getHref()).isEqualTo("/api/vorgangs?page=0&limit=100&status=" + status.toString());
			}

			@ParameterizedTest
			@EnumSource(mode = Mode.INCLUDE, names = { "NEU", "ANGENOMMEN", "IN_BEARBEITUNG", "BESCHIEDEN", "ABGESCHLOSSEN", "VERWORFEN",
					"ZU_LOESCHEN", "WEITERGELEITET" })
			void shouldCallFormatLinkRelTemplate(VorgangStatus status) {
				viewLinkHandler.buildGetAllVorgaengeByStatus(status, RootViewLinkHandler.VORGAENGE_MY_REL_TEMPLATE);

				verify(viewLinkHandler).formatLinkRelTemplate(RootViewLinkHandler.VORGAENGE_MY_REL_TEMPLATE, status);
			}

			@ParameterizedTest
			@EnumSource(mode = Mode.INCLUDE, names = { "NEU", "ANGENOMMEN", "IN_BEARBEITUNG", "BESCHIEDEN", "ABGESCHLOSSEN", "VERWORFEN",
					"ZU_LOESCHEN", "WEITERGELEITET" })
			void shouldHaveFormattedRelation(VorgangStatus status) {
				var expectedRelation = LoremIpsum.getInstance().getWords(1);
				doReturn(expectedRelation).when(viewLinkHandler).formatLinkRelTemplate(any(), any());

				var link = viewLinkHandler.buildGetAllVorgaengeByStatus(status, RootViewLinkHandler.VORGAENGE_MY_REL_TEMPLATE);

				assertThat(link.getRel()).isEqualTo(LinkRelation.of(expectedRelation));
			}

		}

		@DisplayName("build get all unassigned vorgaenge by status")
		@Nested
		class TestBuildGetAllUnassignedVorgaengeByStatus {

			@ParameterizedTest
			@EnumSource(mode = Mode.INCLUDE, names = { "NEU", "ANGENOMMEN", "IN_BEARBEITUNG", "BESCHIEDEN", "ABGESCHLOSSEN", "VERWORFEN",
					"ZU_LOESCHEN", "WEITERGELEITET" })
			void shouldAddByUnassignedToAndStatus(VorgangStatus status) {
				var link = viewLinkHandler.buildGetAllVorgaengeByAssignedToAndStatus(status, RootViewLinkHandler.VORGAENGE_MY_REL_TEMPLATE,
						UserId.empty());

				assertThat(link.getHref()).isEqualTo("/api/vorgangs?page=0&limit=100&assignedTo=&status=" + status.toString());
			}
		}
	}

	@Nested
	class TestBuildGelAllByUngeleseneNachrichtenLink {

		@Test
		void shouldHaveLinkHref() {
			var link = viewLinkHandler.buildGelAllByUngeleseneNachrichtenLink();

			assertThat(link)
					.extracting(Link::getHref)
					.isEqualTo("/api/vorgangs?page=0&limit=100&nachrichten=ungelesene");
		}

		@Test
		void shouldHaveLinkRel() {
			var link = viewLinkHandler.buildGelAllByUngeleseneNachrichtenLink();

			assertThat(link)
					.extracting(Link::getRel)
					.extracting(LinkRelation::value)
					.isEqualTo(RootViewLinkHandler.ALL_UNGELESENE_NACHRICHTEN_REL);
		}
	}

	@Nested
	class TestBuildGetAllByAssignedToAndUngeleseneNachrichten {

		@Nested
		class AssignTo {

			@Test
			void shouldHaveLink() {
				var link = viewLinkHandler.buildGetAllByAssignedToAndUngeleseneNachrichten(UserProfileTestFactory.ID,
						RootViewLinkHandler.MY_UNGELESENE_NACHRICHTEN_REL);

				assertThat(link)
						.extracting(Link::getHref)
						.isEqualTo("/api/vorgangs?page=0&limit=100&assignedTo=" + UserProfileTestFactory.ID + "&nachrichten=ungelesene");
			}

			@Test
			void shouldHaveLinkRel() {
				var link = viewLinkHandler.buildGetAllByAssignedToAndUngeleseneNachrichten(UserProfileTestFactory.ID,
						RootViewLinkHandler.MY_UNGELESENE_NACHRICHTEN_REL);

				assertThat(link)
						.extracting(Link::getRel)
						.extracting(LinkRelation::value)
						.isEqualTo(RootViewLinkHandler.MY_UNGELESENE_NACHRICHTEN_REL);
			}
		}

		@Nested
		class Unassigned {

			@Test
			void shouldHaveLink() {
				var link = viewLinkHandler.buildGetAllByAssignedToAndUngeleseneNachrichten(UserId.empty(),
						RootViewLinkHandler.UNASSIGNED_UNGELESENE_NACHRICHTEN_REL);

				assertThat(link)
						.extracting(Link::getHref)
						.isEqualTo("/api/vorgangs?page=0&limit=100&assignedTo=&nachrichten=ungelesene");
			}

			@Test
			void shouldHaveLinkRel() {
				var link = viewLinkHandler.buildGetAllByAssignedToAndUngeleseneNachrichten(UserId.empty(),
						RootViewLinkHandler.UNASSIGNED_UNGELESENE_NACHRICHTEN_REL);

				assertThat(link)
						.extracting(Link::getRel)
						.extracting(LinkRelation::value)
						.isEqualTo(RootViewLinkHandler.UNASSIGNED_UNGELESENE_NACHRICHTEN_REL);
			}

		}

	}

	@Nested
	class TestAddGetByUngeleseneNachrichtenLinks {

		private Root rootResource = RootTestFactory.create();
		private final EntityModel<Root> model = EntityModel.of(rootResource);
		private final Link linkAllUngeleseneNachrichten = Link.of(LoremIpsum.getInstance().getUrl());
		private final Link linkMyUngeleseneNachrichten = Link.of(LoremIpsum.getInstance().getUrl());
		private final Link linkUnassignedUngeleseneNachrichten = Link.of(LoremIpsum.getInstance().getUrl());

		@BeforeEach
		void setUp() {
			doReturn(linkAllUngeleseneNachrichten).when(viewLinkHandler).buildGelAllByUngeleseneNachrichtenLink();
			doReturn(linkUnassignedUngeleseneNachrichten).when(viewLinkHandler)
					.buildGetAllByAssignedToAndUngeleseneNachrichten(UserId.empty(), RootViewLinkHandler.UNASSIGNED_UNGELESENE_NACHRICHTEN_REL);
		}

		@Test
		void shouldBuildLinkForUngeleseneNachrichten() {
			viewLinkHandler.addGetByUngeleseneNachrichtenLinks(model, Optional.of(UserProfileTestFactory.ID));

			verify(viewLinkHandler).buildGelAllByUngeleseneNachrichtenLink();
		}

		@Test
		void shouldAddLinkForUngeleseneNachrichten() {
			var model = EntityModel.of(rootResource);

			viewLinkHandler.addGetByUngeleseneNachrichtenLinks(model, Optional.of(UserProfileTestFactory.ID));

			assertThat(model.getLinks()).contains(linkAllUngeleseneNachrichten);
		}

		@Test
		void shouldBuildLinkForUnassignedUngeleseneNachrichten() {
			viewLinkHandler.addGetByUngeleseneNachrichtenLinks(model, Optional.empty());

			verify(viewLinkHandler).buildGetAllByAssignedToAndUngeleseneNachrichten(UserId.empty(),
					RootViewLinkHandler.UNASSIGNED_UNGELESENE_NACHRICHTEN_REL);
		}

		@Test
		void shouldAddLinkForUnassignedUngeleseneNachrichten() {
			var model = EntityModel.of(rootResource);

			viewLinkHandler.addGetByUngeleseneNachrichtenLinks(model, Optional.empty());

			assertThat(model.getLinks()).contains(linkUnassignedUngeleseneNachrichten);
		}

		@Nested
		class UsedIdExists {

			@BeforeEach
			void setUp() {
				doReturn(linkMyUngeleseneNachrichten).when(viewLinkHandler).buildGetAllByAssignedToAndUngeleseneNachrichten(UserProfileTestFactory.ID,
						RootViewLinkHandler.MY_UNGELESENE_NACHRICHTEN_REL);
			}

			@Test
			void shouldBuildLinkForMyUngeleseneNachrichten() {
				viewLinkHandler.addGetByUngeleseneNachrichtenLinks(model, Optional.of(UserProfileTestFactory.ID));

				verify(viewLinkHandler).buildGetAllByAssignedToAndUngeleseneNachrichten(UserProfileTestFactory.ID,
						RootViewLinkHandler.MY_UNGELESENE_NACHRICHTEN_REL);
			}

			@Test
			void shouldAddLinkForMyUngeleseneNachrichten() {
				var model = EntityModel.of(rootResource);

				viewLinkHandler.addGetByUngeleseneNachrichtenLinks(model, Optional.of(UserProfileTestFactory.ID));

				assertThat(model.getLinks()).contains(linkMyUngeleseneNachrichten);
			}
		}

		@Nested
		class UserIdEmpty {

			@Test
			void shouldNotBuildLinkForMyUngeleseneNachrichten() {
				viewLinkHandler.addGetByUngeleseneNachrichtenLinks(model, Optional.empty());

				verify(viewLinkHandler, never()).buildGetAllByAssignedToAndUngeleseneNachrichten(UserProfileTestFactory.ID,
						RootViewLinkHandler.MY_UNGELESENE_NACHRICHTEN_REL);
			}
		}

	}

	@Nested
	class TestAddVorgangStatusLinks {

		private final EntityModel<Root> model = EntityModel.of(RootTestFactory.create());
		private final List<VorgangStatus> vorgangStatuses = List.of(VorgangStatus.NEU, VorgangStatus.BESCHIEDEN);
		private final List<Link> links = LinkTestFactory.createList(vorgangStatuses.size());

		@BeforeEach
		void mock() {
			doReturn(vorgangStatuses.stream()).when(viewLinkHandler).getVorgangStatuses();
			doAnswer(AdditionalAnswers.returnsElementsOf(links)).when(viewLinkHandler).buildGetAllVorgaengeByStatus(any(), any());
		}

		@Test
		void shouldCallGetVorgangStatuses() {
			viewLinkHandler.addVorgangStatusLinks(model, RootViewLinkHandler.VORGAENGE_ALL_REL_TEMPLATE);

			verify(viewLinkHandler).getVorgangStatuses();
		}

		@Test
		void shouldCallBuildGetAllVorgaengeByStatus() {
			viewLinkHandler.addVorgangStatusLinks(model, RootViewLinkHandler.VORGAENGE_ALL_REL_TEMPLATE);

			vorgangStatuses
					.forEach(status -> verify(viewLinkHandler).buildGetAllVorgaengeByStatus(status, RootViewLinkHandler.VORGAENGE_ALL_REL_TEMPLATE));
		}

		@Test
		void shouldAddLinksToModel() {
			viewLinkHandler.addVorgangStatusLinks(model, RootViewLinkHandler.VORGAENGE_ALL_REL_TEMPLATE);

			assertThat(model.getLinks()).containsExactlyInAnyOrderElementsOf(links);
		}
	}

	@Nested
	class TestGetVorgangStatuses {

		@Test
		void shouldCallFeatureToggleProperties() {
			viewLinkHandler.getVorgangStatuses().toList();

			verify(featureToggleProperties).isForwardByOzgCloudEnabled();
		}

		@Nested
		class TestOnIsForwardByOzgCloudEnabled {

			@BeforeEach
			void mock() {
				when(featureToggleProperties.isForwardByOzgCloudEnabled()).thenReturn(true);
			}

			@Test
			void shouldReturnStream() {
				var expectedStatuses = List.of(VorgangStatus.NEU, VorgangStatus.ANGENOMMEN, VorgangStatus.IN_BEARBEITUNG,
						VorgangStatus.BESCHIEDEN, VorgangStatus.ABGESCHLOSSEN, VorgangStatus.VERWORFEN, VorgangStatus.ZU_LOESCHEN,
						VorgangStatus.WEITERGELEITET);

				var stream = viewLinkHandler.getVorgangStatuses();

				assertThat(stream).containsExactlyInAnyOrderElementsOf(expectedStatuses);
			}
		}

		@Nested
		class TestOnIsNotForwardByOzgCloudEnabled {

			@BeforeEach
			void mock() {
				when(featureToggleProperties.isForwardByOzgCloudEnabled()).thenReturn(false);
			}

			@Test
			void shouldReturnStream() {
				var expectedStatuses = List.of(VorgangStatus.NEU, VorgangStatus.ANGENOMMEN, VorgangStatus.IN_BEARBEITUNG,
						VorgangStatus.BESCHIEDEN, VorgangStatus.ABGESCHLOSSEN, VorgangStatus.VERWORFEN, VorgangStatus.ZU_LOESCHEN);

				var stream = viewLinkHandler.getVorgangStatuses();

				assertThat(stream).containsExactlyInAnyOrderElementsOf(expectedStatuses);
			}
		}
	}
}
