/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.alfa.bescheid;

import static de.ozgcloud.alfa.common.command.CommandController.*;
import static org.assertj.core.api.Assertions.*;
import static org.mockito.Mockito.*;
import static org.springframework.test.web.servlet.request.MockMvcRequestBuilders.*;
import static org.springframework.test.web.servlet.result.MockMvcResultMatchers.*;

import java.util.Optional;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.ArgumentCaptor;
import org.mockito.Captor;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;
import org.springframework.http.MediaType;
import org.springframework.test.web.servlet.MockMvc;
import org.springframework.test.web.servlet.ResultActions;
import org.springframework.test.web.servlet.setup.MockMvcBuilders;

import de.ozgcloud.alfa.common.command.Command;
import de.ozgcloud.alfa.common.command.CommandOrder;
import de.ozgcloud.alfa.common.command.CommandService;
import de.ozgcloud.alfa.common.command.CommandTestFactory;
import de.ozgcloud.alfa.common.command.CreateCommand;
import de.ozgcloud.alfa.vorgang.VorgangHeaderTestFactory;
import de.ozgcloud.common.errorhandling.TechnicalException;
import lombok.SneakyThrows;

class BescheidSendenCommandControllerTest {

	@Spy
	@InjectMocks
	private BescheidCommandController controller;

	@Mock
	private BescheidService bescheidService;
	@Mock
	private CommandService commandService;

	@Mock
	private BescheidSendenCommandValidator bescheidSendenCommandValidator;

	private MockMvc mockMvc;

	@BeforeEach
	void setUp() {
		mockMvc = MockMvcBuilders.standaloneSetup(controller).build();
	}

	@Nested
	class TestCreateCommand {

		private final Command command = CommandTestFactory.create();

		@Captor
		private ArgumentCaptor<CreateCommand> createCommandArgumentCaptor;

		@BeforeEach
		void setUp() {
			doNothing().when(controller).validate(eq(VorgangHeaderTestFactory.ID), any(CreateCommand.class));
			when(commandService.createCommand(any(CreateCommand.class), eq(BescheidTestFactory.VERSION))).thenReturn(command);
		}

		@SneakyThrows
		@Test
		void shouldValidate() {
			doRequest();

			verify(controller).validate(eq(VorgangHeaderTestFactory.ID), any(CreateCommand.class));
		}

		@SneakyThrows
		@Test
		void shouldSetVorgangIdOnCreateCommand() {
			doRequest();

			verify(controller).validate(eq(VorgangHeaderTestFactory.ID), createCommandArgumentCaptor.capture());
			assertThat(createCommandArgumentCaptor.getValue().getVorgangId()).isEqualTo(VorgangHeaderTestFactory.ID);
		}

		@SneakyThrows
		@Test
		void shouldSetRelationIdOnCreateCommand() {
			doRequest();

			verify(controller).validate(eq(VorgangHeaderTestFactory.ID), createCommandArgumentCaptor.capture());
			assertThat(createCommandArgumentCaptor.getValue().getRelationId()).isEqualTo(BescheidTestFactory.ID);
		}

		@SneakyThrows
		@Test
		void shouldCallCommandService() {
			doRequest();

			verify(commandService).createCommand(any(CreateCommand.class), eq(BescheidTestFactory.VERSION));
		}

		@SneakyThrows
		@Test
		void shouldReturnCreated() {
			var response = doRequest();

			response.andExpect(status().isCreated());
		}

		@SneakyThrows
		@Test
		void shouldReturnCommand() {
			doRequest()
					.andExpect(header().stringValues("location", "http://localhost" + COMMANDS_PATH + "/" + CommandTestFactory.ID));
		}

		@SneakyThrows
		private ResultActions doRequest() {
			return mockMvc.perform(
					post(BescheidCommandController.PATH,
							VorgangHeaderTestFactory.ID,
							BescheidTestFactory.ID,
							BescheidTestFactory.VERSION)
							.content(CommandTestFactory.buildCreateCommandContent(CommandOrder.SEND_BESCHEID))
							.contentType(MediaType.APPLICATION_JSON));
		}

	}

	@Nested
	class TestValidate {

		private final CreateCommand createCommand = CommandTestFactory.createCreateCommand();
		private final Bescheid bescheid = BescheidTestFactory.create();

		@Captor
		private ArgumentCaptor<CreateCommand> commandArgumentCaptor;

		@Nested
		class WhenBescheidExists {

			@BeforeEach
			void setUp() {
				when(bescheidService.getBescheidDraft(VorgangHeaderTestFactory.ID)).thenReturn(Optional.of(bescheid));
				doNothing().when(bescheidSendenCommandValidator).validate(any(CreateCommand.class));
			}

			@Test
			void shouldGetBescheidDraft() {
				controller.validate(VorgangHeaderTestFactory.ID, createCommand);

				verify(bescheidService).getBescheidDraft(VorgangHeaderTestFactory.ID);
			}

			@Test
			void shouldValidate() {
				controller.validate(VorgangHeaderTestFactory.ID, createCommand);

				verify(bescheidSendenCommandValidator).validate(any(CreateCommand.class));
			}

			@Test
			void shouldAddBescheidToCommand() {
				controller.validate(VorgangHeaderTestFactory.ID, createCommand);

				verify(bescheidSendenCommandValidator).validate(commandArgumentCaptor.capture());
				assertThat(commandArgumentCaptor.getValue()).usingRecursiveComparison().isEqualTo(createCommand.toBuilder().body(bescheid).build());
			}
		}

		@Nested
		class WhenBescheidNotExists {

			@Test
			void shouldThrowTechnicalException() {
				when(bescheidService.getBescheidDraft(VorgangHeaderTestFactory.ID)).thenReturn(Optional.empty());

				assertThatThrownBy(() -> controller.validate(VorgangHeaderTestFactory.ID, createCommand)).isInstanceOf(TechnicalException.class);
			}
		}
	}

}