/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.alfa.collaboration;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.Mockito.*;

import java.util.stream.Stream;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;
import org.springframework.hateoas.CollectionModel;
import org.springframework.hateoas.EntityModel;
import org.springframework.hateoas.IanaLinkRelations;
import org.springframework.hateoas.Link;
import org.springframework.hateoas.UriTemplate;

import de.ozgcloud.alfa.collaboration.CollaborationController.CollaborationByVorgangController;
import de.ozgcloud.alfa.common.command.CommandController.CommandByRelationController;
import de.ozgcloud.alfa.vorgang.VorgangController;
import de.ozgcloud.alfa.vorgang.VorgangHeaderTestFactory;
import de.ozgcloud.alfa.vorgang.VorgangWithEingang;
import de.ozgcloud.alfa.vorgang.VorgangWithEingangTestFactory;

class CollaborationModelAssemblerTest {

	@Spy
	@InjectMocks
	private CollaborationModelAssembler assembler;

	@Mock
	private VorgangController vorgangController;

	@Nested
	class TestToModel {

		@Nested
		class OnOrganisationsEinheitCollaboration {

			private final OrganisationsEinheitCollaboration collaboration = OrganisationsEinheitCollaborationTestFactory.create();

			@Test
			void shouldHaveContent() {
				var entityModel = callAssembler();

				assertThat(entityModel.getContent()).isEqualTo(collaboration);
			}

			private EntityModel<Collaboration> callAssembler() {
				return assembler.toModel(collaboration);
			}
		}

		@Nested
		class OnFachstelleCollaboration {

			private final FachstelleCollaboration collaboration = FachstelleCollaborationTestFactory.create();

			@Test
			void shouldHaveContent() {
				var entityModel = callAssembler();

				assertThat(entityModel.getContent()).isEqualTo(collaboration);
			}

			private EntityModel<Collaboration> callAssembler() {
				return assembler.toModel(collaboration);
			}
		}
	}

	@Nested
	class TestToCollectionModel {

		private final Collaboration collaboration = FachstelleCollaborationTestFactory.create();
		private final VorgangWithEingang vorgang = VorgangWithEingangTestFactory.create();

		@BeforeEach
		void mockVorgangController() {
			when(vorgangController.getVorgang(VorgangHeaderTestFactory.ID)).thenReturn(vorgang);
		}

		@Nested
		class OnNonEmptyCollaborationList {

			@Test
			void shouldGetVorgang() {
				callAssembler();

				verify(vorgangController).getVorgang(VorgangHeaderTestFactory.ID);
			}

			@Test
			void shouldCallToModel() {
				callAssembler();

				verify(assembler).toModel(collaboration);
			}

			@Test
			void shouldHaveEntityModel() {
				var entityModel = EntityModel.of(collaboration);
				doReturn(entityModel).when(assembler).toModel(collaboration);

				var collectionModel = callAssembler();

				assertThat(collectionModel.getContent()).containsExactly(entityModel);
			}

			@Test
			void shouldHaveSelfLink() {
				var collectionModel = callAssembler();

				assertThat(collectionModel.getLink(IanaLinkRelations.SELF_VALUE)).get().extracting(Link::getHref)
						.isEqualTo(CollaborationByVorgangController.PATH + "/" + VorgangHeaderTestFactory.ID + "/collaborations");
			}

			@Test
			void shouldNotHaveCreateCollaborationRequestLink() {
				var collectionModel = callAssembler();

				assertThat(collectionModel.getLink(CollaborationModelAssembler.REL_CREATE_COLLABORATION_REQUEST)).isEmpty();
			}

			private CollectionModel<EntityModel<Collaboration>> callAssembler() {
				return assembler.toCollectionModel(Stream.of(collaboration), VorgangHeaderTestFactory.ID);
			}
		}

		@Nested
		class OnEmptyCollaborationList {

			@Test
			void shouldGetVorgang() {
				callAssembler();

				verify(vorgangController).getVorgang(VorgangHeaderTestFactory.ID);
			}

			@Test
			void shouldHaveEmptyContent() {
				var collectionModel = callAssembler();

				assertThat(collectionModel.getContent()).isEmpty();
			}

			@Test
			void shouldHaveSelfLink() {
				var collectionModel = callAssembler();

				assertThat(collectionModel.getLink(IanaLinkRelations.SELF_VALUE)).get().extracting(Link::getHref)
						.isEqualTo(CollaborationByVorgangController.PATH + "/" + VorgangHeaderTestFactory.ID + "/collaborations");
			}

			@Nested
			class OnVorgangIsEditable {

				@Mock
				private Link createCollaborationRequestLink;

				@BeforeEach
				void mock() {
					when(vorgangController.isEditable(vorgang)).thenReturn(true);
					doReturn(createCollaborationRequestLink).when(assembler).buildCreateCollaborationRequestLink(vorgang);
				}

				@Test
				void shouldHaveCreateCollaborationRequestLink() {
					var collectionModel = callAssembler();

					assertThat(collectionModel.getLinks()).contains(createCollaborationRequestLink);
				}
			}

			@Nested
			class OnVorgangIsNotEditable {

				@BeforeEach
				void mock() {
					when(vorgangController.isEditable(vorgang)).thenReturn(false);
				}

				@Test
				void shouldNotHaveCreateCollaborationRequestLink() {
					var collectionModel = callAssembler();

					assertThat(collectionModel.getLink(CollaborationModelAssembler.REL_CREATE_COLLABORATION_REQUEST)).isEmpty();
				}
			}

			private CollectionModel<EntityModel<Collaboration>> callAssembler() {
				return assembler.toCollectionModel(Stream.empty(), VorgangHeaderTestFactory.ID);
			}
		}
	}

	@Nested
	class TestBuildCreateCollaborationRequestLink {

		private final VorgangWithEingang vorgang = VorgangWithEingangTestFactory.create();

		@Test
		void shouldHaveHrefToCreateCommand() {
			var expectedHref = UriTemplate.of(CommandByRelationController.COMMAND_BY_RELATION_PATH)
					.expand(VorgangHeaderTestFactory.ID, VorgangHeaderTestFactory.ID, VorgangHeaderTestFactory.VERSION).toString();

			var link = callAssembler();

			assertThat(link).extracting(Link::getHref).isEqualTo(expectedHref);
		}

		@Test
		void shouldHaveCreateCollaborationRequestRelation() {
			var link = callAssembler();

			assertThat(link).extracting(Link::getRel).isEqualTo(CollaborationModelAssembler.REL_CREATE_COLLABORATION_REQUEST);
		}

		private Link callAssembler() {
			return assembler.buildCreateCollaborationRequestLink(vorgang);
		}
	}
}
