/*
 * Copyright (C) 2025 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.alfa.common;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.lang.reflect.Field;
import java.util.Arrays;
import java.util.List;
import java.util.stream.Stream;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.NullAndEmptySource;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;
import org.springframework.hateoas.EntityModel;
import org.springframework.hateoas.Link;

import com.thedeanda.lorem.LoremIpsum;

import de.ozgcloud.alfa.common.user.UserId;
import de.ozgcloud.alfa.common.user.UserManagerUrlProvider;
import lombok.SneakyThrows;

class LinkedResourceProcessorTest {

	@Spy
	@InjectMocks
	private LinkedResourceProcessor<TestEntity> processor;

	@Mock
	private UserManagerUrlProvider userManagerUrlProvider;

	@Nested
	class TestProcess {

		private final TestEntity entity = TestEntityTestFactory.create();
		private final EntityModel<TestEntity> model = EntityModel.of(entity);

		@BeforeEach
		void mock() {
			doNothing().when(processor).addResourceLinks(any());
			doNothing().when(processor).addUserProfileLinks(any());
		}

		@Test
		void shouldReturnSameEntityModel() {
			var result = processor.process(model);

			assertThat(result).isSameAs(model);
		}

		@Test
		void shouldCallAddResourceLinks() {
			processor.process(model);

			verify(processor).addResourceLinks(model);
		}

		@Test
		void shouldCallAddUserProfileLinks() {
			processor.process(model);

			verify(processor).addUserProfileLinks(model);
		}
	}

	@Nested
	class TestAddResourceLinks {

		private final TestEntity entity = TestEntityTestFactory.create();
		private final EntityModel<TestEntity> model = EntityModel.of(entity);

		@Mock
		private Field field;

		@BeforeEach
		void mock() {
			doReturn(Stream.of(field)).when(processor).getAnnotatedFields(any(), any());
		}

		@Test
		void shouldGetAnnotatedFieldsWithLinkedResourceAnnotation() {
			doReturn(false).when(processor).shouldAddLink(any(), any());

			processor.addResourceLinks(model);

			verify(processor).getAnnotatedFields(LinkedResource.class, entity);
		}

		@Test
		void shouldCallShouldAddLink() {
			doReturn(false).when(processor).shouldAddLink(any(), any());

			processor.addResourceLinks(model);

			verify(processor).shouldAddLink(model, field);
		}

		@Test
		void shouldCallAddLinkForLinkedResourceField() {
			doReturn(true).when(processor).shouldAddLink(any(), any());

			processor.addResourceLinks(model);

			verify(processor).addLinkForLinkedResourceField(model, field);
		}

		@Test
		void shouldNotCallAddLinkForLinkedResourceField() {
			doReturn(false).when(processor).shouldAddLink(any(), any());

			processor.addResourceLinks(model);

			verify(processor, never()).addLinkForLinkedResourceField(any(), any());
		}
	}

	@Nested
	class TestAddUserProfileLinks {

		private final TestEntity entity = TestEntityTestFactory.create();
		private final EntityModel<TestEntity> model = EntityModel.of(entity);

		@Mock
		private Field field;

		@BeforeEach
		void mock() {
			doReturn(Stream.of(field)).when(processor).getAnnotatedFields(any(), any());
		}

		@Test
		void shouldGetAnnotatedFieldsWithLinkedUserProfileResourceAnnotation() {
			doReturn(false).when(processor).shouldAddLink(any(), any());

			processor.addUserProfileLinks(model);

			verify(processor).getAnnotatedFields(LinkedUserProfileResource.class, entity);
		}

		@Test
		void shouldCallShouldAddLink() {
			doReturn(false).when(processor).shouldAddLink(any(), any());

			processor.addUserProfileLinks(model);

			verify(processor).shouldAddLink(model, field);
		}

		@Test
		void shouldCallAddLinkForLinkedUserProfileResourceField() {
			doReturn(true).when(processor).shouldAddLink(any(), any());

			processor.addUserProfileLinks(model);

			verify(processor).addLinkForLinkedUserProfileResourceField(model, field);
		}

		@Test
		void shouldNotCallAddLinkForLinkedUserProfileResourceField() {
			doReturn(false).when(processor).shouldAddLink(any(), any());

			processor.addUserProfileLinks(model);

			verify(processor, never()).addLinkForLinkedUserProfileResourceField(any(), any());
		}
	}

	@Nested
	class TestGetAnnotatedFields {

		@Test
		void shouldReturnEmptyStreamIfContentIsNull() {
			var result = processor.getAnnotatedFields(LinkedResource.class, null);

			assertThat(result).isEmpty();
		}

		@Test
		void shouldReturnAllFieldsWithLinkedResourceAnnotation() {
			var entity = TestEntityTestFactory.create();
			var expectedFields = List.of(getField("linkedResource"), getField("testId"));

			var result = processor.getAnnotatedFields(LinkedResource.class, entity);

			assertThat(result).containsExactlyInAnyOrderElementsOf(expectedFields);
		}

		@Test
		void shouldReturnAllFieldsWithLinkedUserProfileResource() {
			var entity = TestEntityTestFactory.create();
			var expectedFields = List.of(getField("user"), getField("differentUserId"));

			var result = processor.getAnnotatedFields(LinkedUserProfileResource.class, entity);

			assertThat(result).containsExactlyInAnyOrderElementsOf(expectedFields);
		}

		@SneakyThrows
		private Field getField(String fieldName) {
			return TestEntity.class.getDeclaredField(fieldName);
		}
	}

	@Nested
	class TestShouldAddLink {

		private final TestEntity entity = TestEntityTestFactory.create();
		private final EntityModel<TestEntity> model = EntityModel.of(entity);

		@Test
		void shouldReturnFalseIfFieldIsArray() {
			var field = getField("arrayField");

			var result = processor.shouldAddLink(model, field);

			assertThat(result).isFalse();
		}

		@Test
		void shouldReturnFalseIfFieldIsCollection() {
			var field = getField("collectionField");

			var result = processor.shouldAddLink(model, field);

			assertThat(result).isFalse();
		}

		@Test
		void shouldReturnFalseIfFieldHasLink() {
			var field = getField("linkedResource");
			model.add(Link.of(LoremIpsum.getInstance().getUrl()).withRel(TestIdController.LINKED_RESOURCE_REL));

			var result = processor.shouldAddLink(model, field);

			assertThat(result).isFalse();
		}

		@Test
		void shouldReturnTrueOtherwise() {
			var fieldsToBuildLinksFor = Arrays.stream(TestEntity.class.getDeclaredFields())
					.filter(field -> !(field.getName().equals("collectionField") || field.getName().equals("arrayField")));

			var result = fieldsToBuildLinksFor.map(field -> processor.shouldAddLink(model, field));

			assertThat(result).isNotEmpty().allMatch(value -> value.equals(true));
		}

		@SneakyThrows
		private Field getField(String fieldName) {
			return TestEntity.class.getDeclaredField(fieldName);
		}
	}

	@Nested
	class TestAddLinkForLinkedResourceField {

		@Test
		void shouldAddLinkOfLinkedResource() {
			var model = EntityModel.of(TestEntityTestFactory.create());
			var field = getField("linkedResource");

			processor.addLinkForLinkedResourceField(model, field);

			assertThat(model.getLink(TestIdController.LINKED_RESOURCE_REL).get().getHref())
					.isEqualTo(TestIdController.PATH + "/" + TestEntityTestFactory.LINKED_RESOURCE);
		}

		@Test
		void shouldTrimIdSuffix() {
			var model = EntityModel.of(TestEntityTestFactory.create());
			var field = getField("testId");

			processor.addLinkForLinkedResourceField(model, field);

			assertThat(model.getLink(TestIdController.ID_REL).get().getHref())
					.isEqualTo(TestIdController.PATH + "/" + TestEntityTestFactory.ID);
		}

		@ParameterizedTest
		@NullAndEmptySource
		void shouldNotAddLinkIfFieldValueIsNullOrBlank(String linkedResourceValue) {
			var model = EntityModel.of(TestEntityTestFactory.createBuilder().linkedResource(linkedResourceValue).build());
			var field = getField("linkedResource");

			processor.addLinkForLinkedResourceField(model, field);

			assertThat(model.getLink(TestIdController.LINKED_RESOURCE_REL)).isEmpty();
		}

		@SneakyThrows
		private Field getField(String fieldName) {
			return TestEntity.class.getDeclaredField(fieldName);
		}
	}

	@Nested
	class TestAddLinkForLinkedUserProfileResourceField {

		private final String userProfileTemplate = LoremIpsum.getInstance().getUrl() + "/%s";

		@Test
		void shouldGetUserProfileTemplate() {
			when(userManagerUrlProvider.getUserProfileTemplate()).thenReturn(userProfileTemplate);
			var field = getField("user");

			processor.addLinkForLinkedUserProfileResourceField(EntityModel.of(TestEntityTestFactory.create()), field);

			verify(userManagerUrlProvider).getUserProfileTemplate();
		}

		@Test
		void shouldAddLinkOfLinkedUserProfileResource() {
			when(userManagerUrlProvider.getUserProfileTemplate()).thenReturn(userProfileTemplate);
			var model = EntityModel.of(TestEntityTestFactory.create());
			var field = getField("user");

			processor.addLinkForLinkedUserProfileResourceField(model, field);

			assertThat(model.getLink(TestIdController.USER_REL).get().getHref())
					.isEqualTo(userProfileTemplate.formatted(TestEntityTestFactory.USER));
		}

		@Test
		void shouldTrimIdSuffix() {
			when(userManagerUrlProvider.getUserProfileTemplate()).thenReturn(userProfileTemplate);
			var model = EntityModel.of(TestEntityTestFactory.create());
			var field = getField("differentUserId");

			processor.addLinkForLinkedUserProfileResourceField(model, field);

			assertThat(model.getLink(TestIdController.DIFFERENT_USER_REL).get().getHref())
					.isEqualTo(userProfileTemplate.formatted(TestEntityTestFactory.DIFFERENT_USER_ID));
		}

		@ParameterizedTest
		@NullAndEmptySource
		void shouldNotAddLinkIfFieldValueIsNullOrBlank(String userValue) {
			var model = EntityModel.of(TestEntityTestFactory.createBuilder().user(UserId.from(userValue)).build());
			var field = getField("user");

			processor.addLinkForLinkedUserProfileResourceField(model, field);

			assertThat(model.getLink(TestIdController.USER_REL)).isEmpty();
		}

		@ParameterizedTest
		@NullAndEmptySource
		void shouldNotAddLinkIfUserManagerUrlIsNullOrBlank(String userManagerUrl) {
			when(userManagerUrlProvider.getUserProfileTemplate()).thenReturn(userManagerUrl);
			var model = EntityModel.of(TestEntityTestFactory.create());
			var field = getField("user");

			processor.addLinkForLinkedUserProfileResourceField(model, field);

			assertThat(model.getLink(TestIdController.USER_REL)).isEmpty();
		}

		@SneakyThrows
		private Field getField(String fieldName) {
			return TestEntity.class.getDeclaredField(fieldName);
		}
	}
}
