/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.alfa.export;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.Mockito.*;

import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.EnumSource;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.springframework.hateoas.EntityModel;
import org.springframework.hateoas.Link;
import org.springframework.hateoas.UriTemplate;

import de.ozgcloud.alfa.common.command.CommandController.CommandByRelationController;
import de.ozgcloud.alfa.vorgang.Vorgang.VorgangStatus;
import de.ozgcloud.alfa.vorgang.VorgangController;
import de.ozgcloud.alfa.vorgang.VorgangHeaderTestFactory;
import de.ozgcloud.alfa.vorgang.VorgangWithEingang;
import de.ozgcloud.alfa.vorgang.VorgangWithEingangTestFactory;

class ExportVorgangProcessorTest {

	@InjectMocks
	private ExportVorgangProcessor processor;
	@Mock
	private DmsProperties dmsProperties;
	@Mock
	private VorgangController vorgangController;

	@Nested
	class TestProcess {

		@DisplayName("Export link")
		@Nested
		class TestExportLink {

			@Test
			void shouldAddLink() {
				var vorgang = vorgangInStatus(VorgangStatus.ABGESCHLOSSEN);
				givenVorgangIsEditable(vorgang);

				var model = processor.process(EntityModel.of(vorgang));

				assertThat(model.getLink(ExportVorgangProcessor.REL_EXPORT)).isPresent().get().extracting(Link::getHref)
						.isEqualTo(ExportVorgangController.PATH + "/" + VorgangHeaderTestFactory.ID);
			}

			@ParameterizedTest
			@EnumSource(mode = EnumSource.Mode.EXCLUDE, names = "ABGESCHLOSSEN")
			void shouldNotAddLinkWhenVorgangIsNotAbgeschlossen(VorgangStatus status) {
				var vorgang = vorgangInStatus(status);

				var model = processor.process(EntityModel.of(vorgang));

				assertThat(model.getLink(ExportVorgangProcessor.REL_EXPORT)).isEmpty();
			}

			@Test
			void shouldNotAddLinkWhenVorgangIsNotEditable() {
				var vorgang = vorgangInStatus(VorgangStatus.ABGESCHLOSSEN);
				givenVorgangIsNotEditable(vorgang);

				var model = processor.process(EntityModel.of(vorgang));

				assertThat(model.getLink(ExportVorgangProcessor.REL_EXPORT)).isEmpty();
			}
		}

		@Nested
		class TestArchiveLink {

			@Test
			void shouldAddLink() {
				givenDmsIsEnabled();
				var vorgang = vorgangInStatus(VorgangStatus.ABGESCHLOSSEN);
				givenVorgangIsEditable(vorgang);

				var model = processor.process(EntityModel.of(vorgang));

				assertThat(model.getLink(ExportVorgangProcessor.REL_ARCHIVE)).isPresent().get().extracting(Link::getHref)
						.isEqualTo(UriTemplate.of(CommandByRelationController.COMMAND_BY_RELATION_PATH)
								.expand(VorgangHeaderTestFactory.ID, VorgangHeaderTestFactory.ID, VorgangHeaderTestFactory.VERSION).toString());
			}

			@ParameterizedTest
			@EnumSource(mode = EnumSource.Mode.EXCLUDE, names = "ABGESCHLOSSEN")
			void shouldNotAddLinkWhenVorgangIsNotAbgeschlossen(VorgangStatus status) {
				var vorgang = vorgangInStatus(status);

				var model = processor.process(EntityModel.of(vorgang));

				assertThat(model.getLink(ExportVorgangProcessor.REL_ARCHIVE)).isEmpty();
			}

			@Test
			void shouldNotAddLinkWhenDmsIsNotEnabled() {
				givenDmsIsNotEnabled();
				var vorgang = vorgangInStatus(VorgangStatus.ABGESCHLOSSEN);

				var model = processor.process(EntityModel.of(vorgang));

				assertThat(model.getLink(ExportVorgangProcessor.REL_ARCHIVE)).isEmpty();
			}

			@Test
			void shouldNotAddLinkWhenVorgangIsNotEditable() {
				givenDmsIsEnabled();
				var vorgang = vorgangInStatus(VorgangStatus.ABGESCHLOSSEN);
				givenVorgangIsNotEditable(vorgang);

				var model = processor.process(EntityModel.of(vorgang));

				assertThat(model.getLink(ExportVorgangProcessor.REL_ARCHIVE)).isEmpty();
			}

			void givenDmsIsEnabled() {
				when(dmsProperties.isEnabled()).thenReturn(true);
			}

			void givenDmsIsNotEnabled() {
				when(dmsProperties.isEnabled()).thenReturn(false);
			}
		}

		void givenVorgangIsEditable(VorgangWithEingang vorgang) {
			when(vorgangController.isEditable(vorgang)).thenReturn(true);
		}

		void givenVorgangIsNotEditable(VorgangWithEingang vorgang) {
			when(vorgangController.isEditable(vorgang)).thenReturn(false);
		}

		private VorgangWithEingang vorgangInStatus(VorgangStatus status) {
			return VorgangWithEingangTestFactory.createBuilder().status(status).build();
		}
	}
}