/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.alfa.vorgang.command;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.util.Collections;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.EnumSource;
import org.junit.jupiter.params.provider.EnumSource.Mode;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;
import org.springframework.hateoas.EntityModel;
import org.springframework.hateoas.Link;
import org.springframework.hateoas.LinkRelation;

import de.ozgcloud.alfa.common.command.CommandController;
import de.ozgcloud.alfa.common.user.CurrentUserService;
import de.ozgcloud.alfa.common.user.UserRole;
import de.ozgcloud.alfa.forwarding.ForwardingController;
import de.ozgcloud.alfa.forwarding.ForwardingTestFactory;
import de.ozgcloud.alfa.vorgang.Vorgang.VorgangStatus;
import de.ozgcloud.alfa.vorgang.VorgangHeaderTestFactory;
import de.ozgcloud.alfa.vorgang.VorgangWithEingang;
import de.ozgcloud.alfa.vorgang.VorgangWithEingangTestFactory;

class VorgangWithEingangCommandProcessorTest {

	@Spy
	@InjectMocks
	private VorgangWithEingangCommandProcessor processor;
	@Mock
	private CurrentUserService userService;
	@Mock
	private CommandController commandController;
	@Mock
	private ForwardingController forwardingController;

	private VorgangWithEingang vorgang = VorgangWithEingangTestFactory.create();
	private EntityModel<VorgangWithEingang> vorgangEntityModel = EntityModel.of(vorgang);

	@Test
	void shouldReturnEntityModel() {
		var result = processor.process(vorgangEntityModel);

		assertThat(result).isNotNull();
	}

	@Nested
	class TestAddForwardLink {

		private VorgangWithEingang vorgang = VorgangWithEingangTestFactory.createBuilder().status(VorgangStatus.IN_BEARBEITUNG).build();

		@BeforeEach
		void init() {
			when(userService.hasRole(UserRole.EINHEITLICHER_ANSPRECHPARTNER)).thenReturn(true);
		}

		@DisplayName("should be true for Role EINHEITLICHER_ANSPRECHPARTNER, Status IN_BEARBEITUNG and Forwarding with Error")
		@Test
		void shouldBeTrue() {
			when(forwardingController.findFailedForwardings(anyString())).thenReturn(Collections.singletonList(ForwardingTestFactory.create()));

			var addLink = processor.isForwardingAllowed(vorgang);

			assertThat(addLink).isTrue();
		}

		@Test
		void shouldBeTrueInStatusNeu() {
			var vorgang = VorgangWithEingangTestFactory.createBuilder().status(VorgangStatus.NEU).build();

			var addLink = processor.isForwardingAllowed(vorgang);

			assertThat(addLink).isTrue();
		}

		@Test
		void shouldBeFalseForRoleNotEA() {
			when(userService.hasRole(UserRole.EINHEITLICHER_ANSPRECHPARTNER)).thenReturn(false);

			var addLink = processor.isForwardingAllowed(vorgang);

			assertThat(addLink).isFalse();
		}

		@ParameterizedTest
		@EnumSource(mode = Mode.EXCLUDE, names = { "NEU", "IN_BEARBEITUNG" })
		void shouldBeFalseForVorgangInOtherStatus(VorgangStatus status) {
			var vorgang = VorgangWithEingangTestFactory.createBuilder().status(status).build();

			var addLink = processor.isForwardingAllowed(vorgang);

			assertThat(addLink).isFalse();
		}

		@Test
		void shouldBeFalseIfNoFailedForwarding() {
			when(forwardingController.findFailedForwardings(anyString())).thenReturn(Collections.emptyList());

			var addLink = processor.isForwardingAllowed(vorgang);

			assertThat(addLink).isFalse();
		}

		@Test
		void shouldBeFalseOnStatusWeitergeleitet() {
			VorgangWithEingang vorgang = VorgangWithEingangTestFactory.createBuilder().status(VorgangStatus.WEITERGELEITET).build();

			var addLink = processor.isForwardingAllowed(vorgang);

			assertThat(addLink).isFalse();
		}

	}

	@Nested
	@DisplayName("add 'forward' link")
	class TestForwardLink {

		private final LinkRelation linkRel = VorgangWithEingangCommandProcessor.REL_VORGANG_FORWARD;

		@Test
		void shouldNotBePresent() {
			doReturn(false).when(processor).isForwardingAllowed(any());

			var processed = processor.process(EntityModel.of(VorgangWithEingangTestFactory.create()));

			assertThat(processed.getLink(linkRel)).isEmpty();
		}

		@Test
		void shouldBePresent() {
			doReturn(true).when(processor).isForwardingAllowed(any());

			var processed = processor.process(EntityModel.of(VorgangWithEingangTestFactory.create()));

			assertThat(processed.getLink(linkRel)).isPresent().get()
					.extracting(Link::getHref)
					.isEqualTo("/api/vorgangs/" + VorgangHeaderTestFactory.ID + "/relations/" + VorgangHeaderTestFactory.ID + "/"
							+ VorgangHeaderTestFactory.VERSION + "/commands");
		}

	}

	@Nested
	class TestPendingCommandsLink {

		@Test
		void shouldExists() {
			when(commandController.existsPendingCommands(any())).thenReturn(true);

			var link = processor.process(EntityModel.of(VorgangWithEingangTestFactory.create()))
					.getLink(VorgangWithEingangCommandProcessor.REL_PENDING_COMMANDS);

			assertThat(link).isPresent();
			assertThat(link.get().getHref()).isEqualTo("/api/commands?pending=true&vorgangId=" + VorgangHeaderTestFactory.ID);
		}

		@Test
		void shouldNotExists() {
			when(commandController.existsPendingCommands(any())).thenReturn(false);

			var link = processor.process(EntityModel.of(VorgangWithEingangTestFactory.create()))
					.getLink(VorgangWithEingangCommandProcessor.REL_PENDING_COMMANDS);

			assertThat(link).isNotPresent();
		}

		@DisplayName("should have link with user role " + UserRole.EINHEITLICHER_ANSPRECHPARTNER)
		void shouldExistsByRoleEA() {
			when(commandController.existsPendingCommands(any())).thenReturn(false);

			var link = processor.process(EntityModel.of(VorgangWithEingangTestFactory.create()))
					.getLink(VorgangWithEingangCommandProcessor.REL_PENDING_COMMANDS);

			assertThat(link).isPresent().get().extracting(Link::getHref)
					.isEqualTo("/api/commands?pending=true&vorgangId=" + VorgangHeaderTestFactory.ID);
		}
	}
}